#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "snack.h"
#include "ffv.h"

#ifdef __cplusplus
extern "C" {
#endif

/* Windows magic */

#if defined(__WIN32__)
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN
#define EXPORT(a,b) __declspec(dllexport) a b
BOOL APIENTRY
DllMain(HINSTANCE hInst, DWORD reason, LPVOID reserved)
{
  return TRUE;
}
#else
#define EXPORT(a,b) a b
#endif

int ffvCmd(Sound *s, Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]) {
  if (s->debug > 0) { Snack_WriteLog("Enter ffvCmd\n"); }
  /* option names */
  static CONST84 char *subOptionStrings[] = {
    "-tfra", "-progress", NULL
/*    "-tint", "-text", "-tsep", "-winShape", 
    "-dcorrxformType", "-nOutput", */
  };
  enum subOptions {
    TFRA, PROGRESS
/*    TINT, TEXT, TSET, WINSHAPE, 
    DCORRXFORMTYPE, NOUTPUT,  */
  };
  /* refuse on multichannel audio */
  if (s->nchannels != 1) {
    Tcl_AppendResult(interp, "pitch only works with Mono sounds",
		     (char *) NULL);
    return TCL_ERROR;
  }
  /* remove any dangling progress callbacks */
  if (s->cmdPtr != NULL) {
    Tcl_DecrRefCount(s->cmdPtr);
    s->cmdPtr = NULL;
  }
  
  /* FIXME: interpret parameters and setup corresponding local variables
            instead, we just use the new ffvDEFAULTCREATE instead */
  double tfra  = default_tfra; /* this is ffv_main's DEFAULT_TFRA */

  /* parameter interpretation and setup */
  int arg = 0;
  for (arg = 2; arg < objc; arg += 2) {
    int index;
    if (Tcl_GetIndexFromObj(interp, objv[arg], subOptionStrings, 
                            "option", 0, &index) != TCL_OK) {
      return TCL_ERROR;
    }
    if (arg + 1 == objc) {
      Tcl_AppendResult(interp, "No argument given for ", 
                       subOptionStrings[index], " option", (char *) NULL);
      return TCL_ERROR;
    }
    switch ((enum subOptions) index) {
      case TFRA: {
        if (Tcl_GetDoubleFromObj(interp, objv[arg + 1], &tfra) != TCL_OK) return TCL_ERROR;
        break;
      }
      case PROGRESS: {
        char *str = Tcl_GetStringFromObj(objv[arg + 1], NULL);
        if (strlen(str) > 0) {
          Tcl_IncrRefCount(objv[arg + 1]);
          s->cmdPtr = objv[arg + 1];
        }
        break;
      }
    }
  }
  FfvComputer * ffvComputerPtr;
  int Tsize;
  int nOutput;
  /* the following should be replaced by ffvCREATE with all the parameters */
  ffvCREATEDEFAULT(&Tsize, &nOutput, &ffvComputerPtr); 
  if (s->debug > 0) { ffvPRINT(ffvComputerPtr); }
  
  /* allocate output buffer */
  double* output = (double*) malloc(nOutput * sizeof(double));
  assert(output != (double *) NULL);
  /* number of samples */
  int nAudio = s->length - 1;
  /* step size (in samples) */
  int Tstep = (int) (tfra * (s->samprate) + 0.5);
  assert( Tstep > 0 );
  
  /* convert snack samples (floats) to ffv samples (shorts) */
  float *snack_audio = (float *) malloc(nAudio * sizeof(float));
  signed short int *ffv_audio = (signed short *) malloc(nAudio * sizeof(signed short int));
  Snack_GetSoundData(s, 0, snack_audio, nAudio);
  int i;
  for (i = 0; i < nAudio; i++) {
    ffv_audio[i] = (signed short int) (snack_audio[i]); /* insure proper rounding */
  }
  free((void*) snack_audio);

  /* the to-be-generated list of FFV vectors for every frame */
  Tcl_Obj *list = Tcl_NewListObj(0, NULL);
  
  int begIdx = 0;
  while (begIdx < nAudio) {
    /* Compute number of samples in the current frame. */
    int endIdx = begIdx + Tsize;
    int nFrameAudio = (endIdx > nAudio) ? (nAudio - begIdx) : (endIdx - begIdx);
    
    ffvCOMPUTE(ffvComputerPtr, ffv_audio, begIdx, nFrameAudio, output);
    
    Tcl_Obj* ffvVector = Tcl_NewListObj(0, NULL);
    Tcl_ListObjAppendElement(interp, list, ffvVector);
    int i;
    for (i = 0; i < nOutput; i++) {
      Tcl_ListObjAppendElement(interp, ffvVector,
          Tcl_NewDoubleObj((float) output[i]));
    }
    Snack_ProgressCallback(s->cmdPtr, interp, "FFV", begIdx / nAudio);
    begIdx += Tstep;
  }
  Snack_ProgressCallback(s->cmdPtr, interp, "FFV", 1);

  /* set TCL returned value */
  Tcl_SetObjResult(interp, list);

  /* clean up */
  free((void*) output);
  free((void*) ffv_audio);
  ffvDESTROY(&ffvComputerPtr);

  return TCL_OK;
}

/*
  Initialize the square package and create a new sound command 'square'.
  The syntax is: sndName square
 */

EXPORT(int, Snackffv_Init)(Tcl_Interp *interp)
{
#ifdef USE_TCL_STUBS
  if (Tcl_InitStubs(interp, "8", 0) == NULL) {
    return TCL_ERROR;
  }
#endif

#ifdef USE_TK_STUBS
    if (Tk_InitStubs(interp, "8", 0) == NULL) {
      return TCL_ERROR;
    }
#endif

#ifdef USE_SNACK_STUBS
  if (Snack_InitStubs(interp, "2", 0) == NULL) {
    return TCL_ERROR;
  }
#endif

  if (Tcl_PkgProvide(interp, "snackffv", "1.0") != TCL_OK) {
    return TCL_ERROR;
  }

  Snack_AddSubCmd(SNACK_SOUND_CMD, "ffv", (Snack_CmdProc *) ffvCmd, NULL);

  return TCL_OK;
}

EXPORT(int, Snackffv_SafeInit)(Tcl_Interp *interp)
{
  return Snackffv_Init(interp);
}



#ifdef __cplusplus
}
#endif
